/*
 * Microwindows based SDL video driver implementation.
 *
 * Copyright (c) 2001 by Greg Haerr <greg@centurysoftware.com>
 */

/*
	SDL - Simple DirectMedia Layer
	Copyright (C) 1997, 1998, 1999, 2000  Sam Lantinga

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.

	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Sam Lantinga
	slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_fbvideo.c,v 1.13.2.9 2000/05/29 23:06:52 hercules Exp $";
#endif

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#include "SDL.h"
#include "SDL_error.h"
#include "SDL_video.h"
#include "SDL_mouse.h"
#include "SDL_sysvideo.h"
#include "SDL_pixels_c.h"
#include "SDL_events_c.h"
#include "SDL_fbvideo.h"
#include "SDL_fbmouse_c.h"
#include "SDL_fbevents_c.h"
#include "SDL_fb3dfx.h"
#include "SDL_fbmatrox.h"

/*
 * TODO: clean up unused vars in *.h
 * fix #if FB stuff, save/restore palette
 * handle window movement
 * know whether Microwindows is running on fb or X
 * force GrArea at runtime on fb
 * move conversion blits to Microwindows client lib
 * fix HWSurface routines
 */

/* 
 * Set FORCEGRAREA to 1 to force using GrArea,
 * and all bits will be copied to the server.
 */
#define FORCEGRAREA		0	/* =1 to force using GrArea*/
#define FBCON_DEBUG

#define MWPORTRAIT_GRAREA	8	/* invalid portrait mode when no fb mapping*/

static int PORTRAIT = MWPORTRAIT_GRAREA;/* reset for fb mapped modes*/
static int		double_buffer;	/* set for portrait modes*/
static GR_WINDOW_ID	wid;
static GR_GC_ID		gc;
static unsigned int	BPP, PITCH, XRES, YRES;
static unsigned char *	physpixels;
static unsigned int	physoffset;
static unsigned int	physoffsetX, physoffsetY;
static unsigned char *	offscreen = NULL;

/* Initialization/Query functions */
static int FB_VideoInit(_THIS, SDL_PixelFormat *vformat);
static SDL_Rect **FB_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
static SDL_Surface *FB_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
static int FB_SetColors(_THIS, int firstcolor, int ncolors);
static void FB_VideoQuit(_THIS);

/* Hardware surface functions */
static int FB_InitHWSurfaces(_THIS, caddr_t base, int size);
static void FB_FreeHWSurfaces(_THIS);
static int FB_AllocHWSurface(_THIS, SDL_Surface *surface);
static int FB_LockHWSurface(_THIS, SDL_Surface *surface);
static void FB_UnlockHWSurface(_THIS, SDL_Surface *surface);
static void FB_FreeHWSurface(_THIS, SDL_Surface *surface);
static void FB_WaitVBL(_THIS);
static int FB_FlipHWSurface(_THIS, SDL_Surface *surface);

/* Internal palette functions */
static void FB_SavePalette(_THIS, struct fb_fix_screeninfo *finfo,
                                  struct fb_var_screeninfo *vinfo);
static void FB_RestorePalette(_THIS);

/* FB driver bootstrap functions */

static int FB_Available(void)
{
#if 0
	int console;
	char *fbdev;

	if (!(fbdev = getenv("FRAMEBUFFER")))
		fbdev = "/dev/fb0";
	console = open(fbdev, O_RDWR, 0);
	if (console >= 0)
		close(console);
	return(console >= 0);
	//if (GrOpen() < 0)
		//return 0;
	//GrClose();
#endif
	/* always use this driver, will try fb then GrOpen*/
	return 1;
}

static void FB_DeleteDevice(SDL_VideoDevice *device)
{
	free(device->hidden);
	free(device);
}

static SDL_VideoDevice *FB_CreateDevice(int devindex)
{
	SDL_VideoDevice *this;

	/* Initialize all variables that we clean on shutdown */
	this = (SDL_VideoDevice *)malloc(sizeof(SDL_VideoDevice));
	if ( this ) {
		memset(this, 0, (sizeof *this));
		this->hidden = (struct SDL_PrivateVideoData *)
				malloc((sizeof *this->hidden));
	}
	if ( (this == NULL) || (this->hidden == NULL) ) {
		SDL_OutOfMemory();
		if ( this ) {
			free(this);
		}
		return(0);
	}
	memset(this->hidden, 0, (sizeof *this->hidden));
	wait_vbl = FB_WaitVBL;

	/* Set the function pointers */
	this->VideoInit = FB_VideoInit;
	this->ListModes = FB_ListModes;
	this->SetVideoMode = FB_SetVideoMode;
	this->SetColors = FB_SetColors;
	this->UpdateRects = NULL;
	this->VideoQuit = FB_VideoQuit;
	this->AllocHWSurface = FB_AllocHWSurface;
	this->CheckHWBlit = NULL;
	this->FillHWRect = NULL;
	this->SetHWColorKey = NULL;
	this->SetHWAlpha = NULL;
	this->LockHWSurface = FB_LockHWSurface;
	this->UnlockHWSurface = FB_UnlockHWSurface;
	this->FlipHWSurface = FB_FlipHWSurface;
	this->FreeHWSurface = FB_FreeHWSurface;
	this->SetIcon = NULL;
	this->SetCaption = NULL;
	this->GetWMInfo = NULL;
	this->FreeWMCursor = FB_FreeWMCursor;
	this->CreateWMCursor = FB_CreateWMCursor;
	this->ShowWMCursor = FB_ShowWMCursor;
	this->WarpWMCursor = FB_WarpWMCursor;
	this->InitOSKeymap = FB_InitOSKeymap;
	this->PumpEvents = FB_PumpEvents;

	this->free = FB_DeleteDevice;

	return this;
}

VideoBootStrap FBCON_bootstrap = {
	"fbcon", FB_Available, FB_CreateDevice
};

static int FB_VideoInit(_THIS, SDL_PixelFormat *vformat)
{
	int hasfb;
	GR_SCREEN_INFO sinfo;

	if (GrOpen() < 0) {
		SDL_SetError("Can't open graphics device");
		return(-1);
	}
	GrGetScreenInfo(&sinfo);

	/* Create the hardware surface lock mutex */
	hw_lock = SDL_CreateMutex();
	if ( hw_lock == NULL ) {
		SDL_SetError("Unable to create lock mutex");
		FB_VideoQuit(this);
		return(-1);
	}

#if !FORCEGRAREA
	hasfb = (GrOpenClientFramebuffer() != NULL);
	if (!hasfb) {
		printf("SDL: Can't directly map framebuffer, using GrArea\n");
		PORTRAIT = MWPORTRAIT_GRAREA;
	} else
		PORTRAIT = sinfo.portrait;	/* use server portrait mode*/
#endif
	if (PORTRAIT & (MWPORTRAIT_LEFT|MWPORTRAIT_RIGHT)) {
		XRES = sinfo.rows;	/* reversed coords*/
		YRES = sinfo.cols;
	} else {
		XRES = sinfo.cols;
		YRES = sinfo.rows;
	}
	BPP = sinfo.bpp;
	PITCH = XRES*(BPP/8);

	/* double buffer all modes except fb-mapped non-portrait*/
	double_buffer = PORTRAIT & (MWPORTRAIT_LEFT|MWPORTRAIT_RIGHT|
				MWPORTRAIT_DOWN|MWPORTRAIT_GRAREA);

	/* use shared memory if will call GrArea*/
	if (PORTRAIT & MWPORTRAIT_GRAREA)
		GrReqShmCmds(XRES*YRES*BPP/8);

	vformat->BitsPerPixel = BPP;
	vformat->Rmask = sinfo.rmask;
	vformat->Gmask = sinfo.gmask;
	vformat->Bmask = sinfo.bmask;

	/* FIXME Save hardware palette, if needed */
	//FB_SavePalette(this, &finfo, &vinfo);

	return(0);
}

SDL_Rect **FB_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
	return (SDL_Rect **)-1;	/* allow any width/height for all modes*/
}

/*
 * These conversion blits are a major pain.
 * Mess with them at your own risk ;-)
 */
static void
ConvBlit16(_THIS, unsigned int X, unsigned int Y, unsigned int W,
	unsigned int H, int flip)
{
	unsigned short *ilineptr;
	unsigned short *olineptr;
	unsigned int x, y;

	if (flip == MWPORTRAIT_RIGHT) {
		/* right portrait mode conversion*/
		unsigned int linerem = PITCH/2 - W;
		ilineptr = (unsigned short *)(offscreen + Y*PITCH + X*BPP/8);
		olineptr = (unsigned short *)physpixels;
		for (y=0; y<H; ++y) {
			unsigned int ox =
				(H-1-y + XRES-H - (physoffsetY+Y)) * BPP/8;
			for (x=0; x<W; ++x) {
				unsigned int oy = (x + physoffsetX + X) * PITCH;
				olineptr[(oy + ox)/2] = *ilineptr++;
			}
			ilineptr += linerem;
		}
	} else if (flip == MWPORTRAIT_LEFT) {
		/* left portrait mode conversion*/
		unsigned int linerem = PITCH/2 - W;
		ilineptr = (unsigned short *)(offscreen + Y*PITCH + X*BPP/8);
		olineptr = (unsigned short *)physpixels;
		for (y=0; y<H; ++y) {
			unsigned int ox = (y + physoffsetY + Y) * BPP/8;
			for (x=0; x<W; ++x) {
				unsigned int oy =
				    (W-1-x + YRES-W - (physoffsetX+X)) * PITCH;
				olineptr[(oy + ox)/2] = *ilineptr++;
			}
			ilineptr += linerem;
		}
	} else if (flip == MWPORTRAIT_NONE) {
		/* straight blit, no conversion*/
		unsigned int linerem = PITCH/2;
		unsigned int offset = Y*PITCH + X*BPP/8;
		ilineptr = (unsigned short *)(offscreen + offset);
		olineptr = (unsigned short *)(physpixels + physoffset +
			offset);
		for (y=0; y<H; ++y) {
			memcpy(olineptr, ilineptr, W<<1);
			olineptr += linerem;
			ilineptr += linerem;
		}
	} else {
		/* straight blit, convert pitch, don't draw on framebuffer*/
		unsigned char *buf = malloc(W * H * 2);
		unsigned int linerem = PITCH/2;
		unsigned int offset = Y*PITCH + X*BPP/8;
		ilineptr = (unsigned short *)(offscreen + offset);
		olineptr = (unsigned short *)buf;
		for (y=0; y<H; ++y) {
			memcpy(olineptr, ilineptr, W<<1);
			olineptr += W;
			ilineptr += linerem;
		}

		SDL_mutexP(hw_lock);
		GrArea(wid, gc, X, Y, W, H, buf, MWPF_TRUECOLOR565);
		SDL_mutexV(hw_lock);

		free(buf);
	}
}

#if LATER
static void
ConvBlit24(_THIS, unsigned int X, unsigned int Y, unsigned int W,
	unsigned int H, int flip)
{
	unsigned char *ilineptr;
	unsigned char *olineptr;
	unsigned int x, y;

	if (flip == MWPORTRAIT_RIGHT) {
		/* right portrait mode conversion*/
		unsigned int linerem = PITCH - W*3;
		ilineptr = (unsigned char *)(offscreen + Y*PITCH + X*BPP/8);
		olineptr = (unsigned char *)physpixels;
		for (y=0; y<H; ++y) {
			unsigned int ox =
				(H-1-y + XRES-H - (physoffsetY+Y)) * BPP/8;
			for (x=0; x<W; ++x) {
				unsigned int oy = (x + physoffsetX + X) * PITCH;
				olineptr[(oy + ox)] = *ilineptr++;
				olineptr[(oy + ox + 1)] = *ilineptr++;
				olineptr[(oy + ox + 2)] = *ilineptr++;
			}
			ilineptr += linerem;
		}
	} else if (flip == MWPORTRAIT_LEFT) {
		/* left portrait mode conversion*/
		unsigned int linerem = PITCH - W*3;
		ilineptr = (unsigned char *)(offscreen + Y*PITCH + X*BPP/8);
		olineptr = (unsigned char *)physpixels;
		for (y=0; y<H; ++y) {
			unsigned int ox = (y + physoffsetY + Y) * BPP/8;
			for (x=0; x<W; ++x) {
				unsigned int oy =
				    (W-1-x + YRES-W - (physoffsetX+X)) * PITCH;
				olineptr[(oy + ox)] = *ilineptr++;
				olineptr[(oy + ox + 1)] = *ilineptr++;
				olineptr[(oy + ox + 2)] = *ilineptr++;
			}
			ilineptr += linerem;
		}
	} else if (flip == 0) {
		/* straight blit, no conversion*/
		unsigned int linerem = PITCH;
		unsigned int offset = Y*PITCH + X*BPP/8;
		ilineptr = (unsigned char *)(offscreen + offset);
		olineptr = (unsigned char *)(physpixels + physoffset +
			offset);
		for (y=0; y<H; ++y) {
			memcpy(olineptr, ilineptr, W*3);
			olineptr += linerem;
			ilineptr += linerem;
		}
	} else {
		/* straight blit, convert pitch, don't draw on framebuffer*/
		unsigned char *buf = malloc(W * H * 3);
		unsigned int linerem = PITCH;
		unsigned int offset = Y*PITCH + X*BPP/8;
		ilineptr = (unsigned long *)(offscreen + offset);
		olineptr = (unsigned long *)buf;
		for (y=0; y<H; ++y) {
			memcpy(olineptr, ilineptr, W*3);
			olineptr += W*3;
			ilineptr += linerem;
		}

		SDL_mutexP(hw_lock);
		GrArea(wid, gc, X, Y, W, H, buf, MWPF_TRUECOLOR888);
		SDL_mutexV(hw_lock);

		free(buf);
	}
}
#endif

static void
ConvBlit32(_THIS, unsigned int X, unsigned int Y, unsigned int W,
	unsigned int H, int flip)
{
	unsigned long *ilineptr;
	unsigned long *olineptr;
	unsigned int x, y;

	if (flip == MWPORTRAIT_RIGHT) {
		/* right portrait mode conversion*/
		unsigned int linerem = PITCH/4 - W;
		ilineptr = (unsigned long *)(offscreen + Y*PITCH + X*BPP/8);
		olineptr = (unsigned long *)physpixels;
		for (y=0; y<H; ++y) {
			unsigned int ox =
				(H-1-y + XRES-H - (physoffsetY+Y)) * BPP/8;
			for (x=0; x<W; ++x) {
				unsigned int oy = (x + physoffsetX + X) * PITCH;
				olineptr[(oy + ox)/4] = *ilineptr++;
			}
			ilineptr += linerem;
		}
	} else if (flip == MWPORTRAIT_LEFT) {
		/* left portrait mode conversion*/
		unsigned int linerem = PITCH/4 - W;
		ilineptr = (unsigned long *)(offscreen + Y*PITCH + X*BPP/8);
		olineptr = (unsigned long *)physpixels;
		for (y=0; y<H; ++y) {
			unsigned int ox = (y + physoffsetY + Y) * BPP/8;
			for (x=0; x<W; ++x) {
				unsigned int oy =
				    (W-1-x + YRES-W - (physoffsetX+X)) * PITCH;
				olineptr[(oy + ox)/4] = *ilineptr++;
			}
			ilineptr += linerem;
		}
	} else if (flip == 0) {
		/* straight blit, no conversion*/
		unsigned int linerem = PITCH/4;
		unsigned int offset = Y*PITCH + X*BPP/8;
		ilineptr = (unsigned long *)(offscreen + offset);
		olineptr = (unsigned long *)(physpixels + physoffset +
			offset);
		for (y=0; y<H; ++y) {
			memcpy(olineptr, ilineptr, W<<2);
			olineptr += linerem;
			ilineptr += linerem;
		}
	} else {
		/* straight blit, convert pitch, don't draw on framebuffer*/
		unsigned char *buf = malloc(W * H * 4);
		unsigned int linerem = PITCH/4;
		unsigned int offset = Y*PITCH + X*BPP/8;
		ilineptr = (unsigned long *)(offscreen + offset);
		olineptr = (unsigned long *)buf;
		for (y=0; y<H; ++y) {
			memcpy(olineptr, ilineptr, W<<2);
			olineptr += W;
			ilineptr += linerem;
		}

		SDL_mutexP(hw_lock);
		GrArea(wid, gc, X, Y, W, H, buf, MWPF_TRUECOLOR0888);
		SDL_mutexV(hw_lock);

		free(buf);
	}
}

static void FB_DirectUpdate(_THIS, int numrects, SDL_Rect *rects)
{
	/* conversion blits and double buffering required on portrait modes*/
	if (!double_buffer)
		return;

	while (--numrects >= 0) {
		switch (BPP) {
		case 16:
			ConvBlit16(this, rects->x, rects->y, rects->w, rects->h,
				PORTRAIT);
			break;
#if LATER
		case 24:
			ConvBlit24(this, rects->x, rects->y, rects->w, rects->h,
				PORTRAIT);
			break;
#endif
		case 32:
			ConvBlit32(this, rects->x, rects->y, rects->w, rects->h,
				PORTRAIT);
			break;
		default:
			fprintf(stderr, "No blitter for %d bpp\n", BPP);
			break;
		}
		++rects;
	}
}

SDL_Surface *FB_SetVideoMode(_THIS, SDL_Surface *current,
				int width, int height, int bpp, Uint32 flags)
{
	int i;
	Uint32 Rmask;
	Uint32 Gmask;
	Uint32 Bmask;
	int size;
	GR_WINDOW_FB_INFO fbinfo;
	GR_SCREEN_INFO sinfo;

fprintf(stderr, "user wants %dx%d %dbpp\n", width, height, bpp);
if (bpp != BPP) fprintf(stderr, "SetVideoMode bad BPP\n");

	/* FIXME Restore the original palette */
	//FB_RestorePalette(this);
	//cache_vinfo = vinfo;

	BPP = bpp;	// may not work FIXME
	PITCH = XRES*(BPP/8);
	GrGetScreenInfo(&sinfo);
	Rmask = sinfo.rmask;
	Gmask = sinfo.gmask;
	Bmask = sinfo.bmask;

	if ( ! SDL_ReallocFormat(current, BPP, Rmask, Gmask, Bmask, 0) )
		return(NULL);

	/* Get the fixed information about the console hardware.
	   This is necessary since finfo.line_length changes.
	 */
	//if ( ioctl(console_fd, FBIOGET_FSCREENINFO, &finfo) < 0 ) {
		//SDL_SetError("Couldn't get console hardware info");
		//return(NULL);
	//}
	/* Save hardware palette, if needed */
	//FB_SavePalette(this, &finfo, &vinfo);

	/* alloc offscreen buffer for moving (must realloc if resized)*/
	if (offscreen)
		free(offscreen);
	size = height*PITCH;
	offscreen = malloc(size);

	wid = GrNewWindowEx(GR_WM_PROPS_APPWINDOW|GR_WM_PROPS_NOBACKGROUND|
		GR_WM_PROPS_NOAUTOMOVE,
		"Microwindows Video",
		GR_ROOT_WINDOW_ID, 0, 0, width, height, BLACK);
	if (!wid) {
		SDL_SetError("Can't create window");
		return(NULL);
	}
	gc = GrNewGC();		/* for possible GrArea()*/
	GrSelectEvents(wid, GR_EVENT_MASK_KEY_DOWN|GR_EVENT_MASK_KEY_UP|
			GR_EVENT_MASK_BUTTON_DOWN|GR_EVENT_MASK_BUTTON_UP|
			GR_EVENT_MASK_CLOSE_REQ|GR_EVENT_MASK_UPDATE|
			0);
	GrMapWindow(wid);
	GrFlush();
	//FIXME: doesn't handle window movement
	GrGetWindowFBInfo(wid, &fbinfo);

	/* Set up the new mode framebuffer */
	current->flags = (SDL_FULLSCREEN|SDL_HWSURFACE);	//FIXME
	current->w = width;
	current->h = height;
	current->pitch = PITCH;
	physpixels = fbinfo.physpixels;
	physoffset = 0;
	physoffsetX = 0;
	physoffsetY = 0;

	/* map to offscreen until window gets UPDATE_MAP event*/
	current->pixels = offscreen;

	/* Update for double-buffering, if we can */
	if ( flags & SDL_DOUBLEBUF ) {
fprintf(stderr, "DOUBLEBUFFER bug\n");
#if 0
		if ( vinfo.yres_virtual == (height*2) ) {
			current->flags |= SDL_DOUBLEBUF;
			flip_page = 0;
			flip_address[0] = (caddr_t)current->pixels;
			flip_address[1] = (caddr_t)current->pixels+
			                           current->h*current->pitch;
			FB_FlipHWSurface(this, current);
		}
#endif
	}

	/* FIXME Set up the information for hardware surfaces */
	//surfaces_mem = (caddr_t)current->pixels + YRES * current->pitch;
	//surfaces_len = (mapped_memlen-(surfaces_mem-mapped_mem));
	//FB_FreeHWSurfaces(this);
	//FB_InitHWSurfaces(this, surfaces_mem, surfaces_len);

	/* Set the update rectangle function */
	this->UpdateRects = FB_DirectUpdate;

	/* We're done */
	return(current);
}

#ifdef FBCON_DEBUG
static void FB_DumpHWSurfaces(_THIS)
{
	vidmem_bucket *bucket;

	printf("Memory left: %d (%d total)\n", surfaces_memleft, surfaces_memtotal);
	printf("\n");
	printf("         Base  Size\n");
	for ( bucket=&surfaces; bucket; bucket=bucket->next ) {
		printf("Bucket:  %p, %d (%s)\n", bucket->base, bucket->size, bucket->used ? "used" : "free");
		if ( bucket->prev ) {
			if ( bucket->base != bucket->prev->base+bucket->prev->size ) {
				printf("Warning, corrupt bucket list! (prev)\n");
			}
		} else {
			if ( bucket != &surfaces ) {
				printf("Warning, corrupt bucket list! (!prev)\n");
			}
		}
		if ( bucket->next ) {
			if ( bucket->next->base != bucket->base+bucket->size ) {
				printf("Warning, corrupt bucket list! (next)\n");
			}
		}
	}
	printf("\n");
}
#endif

static int FB_InitHWSurfaces(_THIS, caddr_t base, int size)
{
	surfaces.prev = NULL;
	surfaces.used = 0;
	surfaces.base = base;
	surfaces.size = size;
	surfaces.next = NULL;
	surfaces_memtotal = size;
	surfaces_memleft = size;
	return(0);
}
static void FB_FreeHWSurfaces(_THIS)
{
	vidmem_bucket *bucket, *freeable;

	bucket = surfaces.next;
	while ( bucket ) {
		freeable = bucket;
		bucket = bucket->next;
		free(freeable);
	}
	surfaces.next = NULL;
}

static int FB_AllocHWSurface(_THIS, SDL_Surface *surface)
{
	vidmem_bucket *bucket;
	int size;
	int extra;

fprintf(stderr, "AllocHWSurface\n");
/* Temporarily, we only allow surfaces the same width as display.
   Some blitters require the pitch between two hardware surfaces
   to be the same.  Others have interesting alignment restrictions.
   Until someone who knows these details looks at the code...
*/
if ( surface->pitch > SDL_VideoSurface->pitch ) {
	SDL_SetError("Surface requested wider than screen");
	return(-1);
}
surface->pitch = SDL_VideoSurface->pitch;
	size = surface->h * surface->pitch;
#ifdef FBCON_DEBUG
	fprintf(stderr, "Allocating bucket of %d bytes\n", size);
#endif

	/* Quick check for available mem */
	if ( size > surfaces_memleft ) {
		SDL_SetError("Not enough video memory");
		return(-1);
	}

	/* Search for an empty bucket big enough */
	for ( bucket=&surfaces; bucket; bucket=bucket->next ) {
		if ( ! bucket->used && (size <= bucket->size) ) {
			break;
		}
	}
	if ( bucket == NULL ) {
		SDL_SetError("Video memory too fragmented");
		return(-1);
	}

	/* Create a new bucket for left-over memory */
	extra = (bucket->size - size);
	if ( extra ) {
		vidmem_bucket *newbucket;

#ifdef FBCON_DEBUG
	fprintf(stderr, "Adding new free bucket of %d bytes\n", extra);
#endif
		newbucket = (vidmem_bucket *)malloc(sizeof(*newbucket));
		if ( newbucket == NULL ) {
			SDL_OutOfMemory();
			return(-1);
		}
		newbucket->prev = bucket;
		newbucket->used = 0;
		newbucket->base = bucket->base+size;
		newbucket->size = extra;
		newbucket->next = bucket->next;
		if ( bucket->next ) {
			bucket->next->prev = newbucket;
		}
		bucket->next = newbucket;
	}

	/* Set the current bucket values and return it! */
	bucket->used = 1;
	bucket->size = size;
#ifdef FBCON_DEBUG
	fprintf(stderr, "Allocated %d bytes at %p\n", bucket->size, bucket->base);
#endif
	surfaces_memleft -= size;
	surface->flags |= SDL_HWSURFACE;
	surface->pixels = bucket->base;
	return(0);
}

static void FB_FreeHWSurface(_THIS, SDL_Surface *surface)
{
	vidmem_bucket *bucket, *freeable;

	/* Look for the bucket in the current list */
	for ( bucket=&surfaces; bucket; bucket=bucket->next ) {
		if ( bucket->base == (caddr_t)surface->pixels ) {
			break;
		}
	}
	if ( (bucket == NULL) || ! bucket->used ) {
		return;
	}

	/* Add the memory back to the total */
#ifdef FBCON_DEBUG
	printf("Freeing bucket of %d bytes\n", bucket->size);
#endif
	surfaces_memleft += bucket->size;

	/* Can we merge the space with surrounding buckets? */
	bucket->used = 0;
	if ( bucket->next && ! bucket->next->used ) {
#ifdef FBCON_DEBUG
	printf("Merging with next bucket, for %d total bytes\n", bucket->size+bucket->next->size);
#endif
		freeable = bucket->next;
		bucket->size += bucket->next->size;
		bucket->next = bucket->next->next;
		if ( bucket->next ) {
			bucket->next->prev = bucket;
		}
		free(freeable);
	}
	if ( bucket->prev && ! bucket->prev->used ) {
#ifdef FBCON_DEBUG
	printf("Merging with previous bucket, for %d total bytes\n", bucket->prev->size+bucket->size);
#endif
		freeable = bucket;
		bucket->prev->size += bucket->size;
		bucket->prev->next = bucket->next;
		if ( bucket->next ) {
			bucket->next->prev = bucket->prev;
		}
		free(freeable);
	}
	surface->pixels = NULL;
}
static int FB_LockHWSurface(_THIS, SDL_Surface *surface)
{
	if ( surface == SDL_VideoSurface ) {
		SDL_mutexP(hw_lock);
	}
	return(0);
}
static void FB_UnlockHWSurface(_THIS, SDL_Surface *surface)
{
	if ( surface == SDL_VideoSurface ) {
		SDL_mutexV(hw_lock);
	}
}

static void FB_WaitVBL(_THIS)
{
#ifdef FBIOWAITRETRACE
	ioctl(console_fd, FBIOWAITRETRACE, 0);
#endif
	return;
}

static int FB_FlipHWSurface(_THIS, SDL_Surface *surface)
{
fprintf(stderr, "FlipHWSurface called: not supported\n"); //FIXME
#if 0
	/* Wait for vertical retrace and then flip display */
	cache_vinfo.yoffset = flip_page*surface->h;
	wait_vbl(this);
	if ( ioctl(console_fd, FBIOPAN_DISPLAY, &cache_vinfo) < 0 ) {
		SDL_SetError("ioctl(FBIOPAN_DISPLAY) failed");
		return(-1);
	}
	flip_page = !flip_page;

	surface->pixels = flip_address[flip_page];
#endif
	return(0);
}

#if FB
void FB_SavePaletteTo(_THIS, int palette_len, __u16 *area)
{
	struct fb_cmap cmap;

return; // leave out for now - move to mwin later
	cmap.start = 0;
	cmap.len = palette_len;
	cmap.red = &area[0*palette_len];
	cmap.green = &area[1*palette_len];
	cmap.blue = &area[2*palette_len];
	cmap.transp = NULL;
	ioctl(console_fd, FBIOGETCMAP, &cmap);
}

void FB_RestorePaletteFrom(_THIS, int palette_len, __u16 *area)
{
	struct fb_cmap cmap;

return; // leave out for now - move to mwin later
	cmap.start = 0;
	cmap.len = palette_len;
	cmap.red = &area[0*palette_len];
	cmap.green = &area[1*palette_len];
	cmap.blue = &area[2*palette_len];
	cmap.transp = NULL;
	ioctl(console_fd, FBIOPUTCMAP, &cmap);
}

static void FB_SavePalette(_THIS, struct fb_fix_screeninfo *finfo,
                                  struct fb_var_screeninfo *vinfo)
{
	int i;

	/* Save hardware palette, if needed */
	if ( finfo->visual == FB_VISUAL_PSEUDOCOLOR ) {
		saved_cmaplen = 1<<vinfo->bits_per_pixel;
		saved_cmap=(__u16 *)malloc(3*saved_cmaplen*sizeof(*saved_cmap));
		if ( saved_cmap != NULL ) {
			FB_SavePaletteTo(this, saved_cmaplen, saved_cmap);
		}
	}

	/* Added support for FB_VISUAL_DIRECTCOLOR.
	   With this mode pixel information is passed through the palette...
	   Neat fading and gamma correction effects can be had by simply
	   fooling around with the palette instead of changing the pixel
	   values themselves... Very neat!

	   Adam Meyerowitz 1/19/2000
	   ameyerow@optonline.com
	*/
	if ( finfo->visual == FB_VISUAL_DIRECTCOLOR ) {
		__u16 new_entries[3*256];

		/* Save the colormap */
		saved_cmaplen = 256;
		saved_cmap=(__u16 *)malloc(3*saved_cmaplen*sizeof(*saved_cmap));
		if ( saved_cmap != NULL ) {
			FB_SavePaletteTo(this, saved_cmaplen, saved_cmap);
		}

		/* Allocate new identity colormap */
		for ( i=0; i<256; ++i ) {
	      		new_entries[(0*256)+i] =
			new_entries[(1*256)+i] =
			new_entries[(2*256)+i] = (i<<8)|i;
		}
		FB_RestorePaletteFrom(this, 256, new_entries);
	}
}

static void FB_RestorePalette(_THIS)
{
	/* Restore the original palette */
	if ( saved_cmap ) {
		FB_RestorePaletteFrom(this, saved_cmaplen, saved_cmap);
		free(saved_cmap);
		saved_cmap = NULL;
	}
}
#endif

static int FB_SetColors(_THIS, int firstcolor, int ncolors)
{
#if FB // FIXME
	SDL_Color *colors;
	int i, which;
	__u16 r[256];
	__u16 g[256];
	__u16 b[256];
	struct fb_cmap cmap;

fprintf(stderr, "SetColors\n");
	/* Set up the colormap */
	colors = this->screen->format->palette->colors;
	for ( i=ncolors-1, which=firstcolor+ncolors-1; i>=0; --i, --which ) {
		r[i] = ((__u16)colors[which].r)<<8;
		g[i] = ((__u16)colors[which].g)<<8;
		b[i] = ((__u16)colors[which].b)<<8;
	}
	cmap.start = firstcolor;
	cmap.len = ncolors;
	cmap.red = r;
	cmap.green = g;
	cmap.blue = b;
	cmap.transp = NULL;

	if ( ioctl(console_fd, FBIOPUTCMAP, &cmap) < 0 ) {
		ncolors = this->screen->format->palette->ncolors;
		cmap.start = 0;
		cmap.len = ncolors;
		if ( ioctl(console_fd, FBIOGETCMAP, &cmap) == 0 ) {
			for ( i=ncolors-1; i>=0; --i ) {
				colors[i].r = (r[i]>>8);
				colors[i].g = (g[i]>>8);
				colors[i].b = (b[i]>>8);
			}
		}
		return(0);
	}
	return(1);
#endif
}

/* Note:  If we are terminated, this could be called in the middle of
   another SDL video routine -- notably UpdateRects.
*/
void FB_VideoQuit(_THIS)
{
	int i, j;

	if ( this->screen ) {
		/* Clear screen and tell SDL not to free the pixels */
		if ( this->screen->pixels ) {
#ifdef __powerpc__	/* SIGBUS when using memset() ?? */
			Uint8 *rowp = (Uint8 *)this->screen->pixels;
			int left = this->screen->pitch*this->screen->h;
			while ( left-- ) { *rowp++ = 0; }
#else
			memset(this->screen->pixels,0,this->screen->h*this->screen->pitch);
#endif
			this->screen->pixels = NULL;
		}
	}

	/* Clear the lock mutex */
	if ( hw_lock != NULL ) {
		SDL_DestroyMutex(hw_lock);
		hw_lock = NULL;
	}

	/* Clean up the memory bucket list */
	FB_FreeHWSurfaces(this);

	GrCloseClientFramebuffer();

	/* Restore the original video mode and palette */
	//FB_RestorePalette(this);

	GrClose();
}

void
FB_SetWindowOffset(_THIS, int x, int y)
{
	SDL_Surface *s = SDL_VideoSurface;

	//FIXME move GrGetWindowFBInfo here?
	if (x < 0) {
		if (!offscreen)
			return;
		s->pixels = offscreen;
		return;
	}
	physoffsetY = y;
	physoffsetX = x;
	physoffset = y*PITCH + x*BPP/8;
	/* if not double-buffering, set physical pixel address to framebuffer*/
	if (!double_buffer)
		s->pixels = physpixels + physoffset;
}
